from web3 import Web3
import web3

from .signature import private_to_stark_key, sign
from .starkex_message import get_limit_order_msg_with_fee, get_transfer_msg

def get_asset_info(type, address):
    if type.lower() == "eth":
        return '0x8322fff2'
    elif type == 'ERC20':
        return '0xf47261b0' + address[2:].rjust(64, '0')
    elif type == 'ERC721':
        return '0x02571792' + address[2:].rjust(64, '0')
    elif type == 'ERC721M':
        return '0xb8b86672' + address[2:].rjust(64, '0')
    else:
        return None

def get_asset_type(typ, address, quantum):
    asset_info = get_asset_info(typ, address)
    asset_type = int(web3.Web3.solidityKeccak(['bytes','uint256'], [asset_info, quantum]).hex(), 16) & 0x03FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    return asset_type

def get_asset_id(typ, address, quantum, token_id):
    asset_id = get_asset_type(typ, address, quantum)
    if typ == 'ERC721':
        asset_id = int(web3.Web3.solidityKeccak(['string','uint256','uint256'], [ 'NFT:',asset_id, token_id]).hex(), 16) & 0x03FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    elif typ == "ERC721M":
        blobHash = int(web3.Web3.solidityKeccak(['bytes'], [ hex(token_id) ]).hex(),16)
        asset_id = int(web3.Web3.solidityKeccak(['string','uint256','uint256'], [ 'MINTABLE:',asset_id, blobHash]).hex(), 16) & 0x0000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
        asset_id = asset_id | 0x400000000000000000000000000000000000000000000000000000000000000
    return asset_id

def get_public_key(private_key):
    return hex(private_to_stark_key(int(private_key,16)))

def get_order_with_fee_signature_local(data):
    private_key = data["sender_private_key"]
    data = {'vault_sell':int(data["vault_id_sell"],10), 'vault_buy':int(data["vault_id_buy"],10), 'amount_sell':int(data["amount_sell"]), 'amount_buy':int(data["amount_buy"]), 'token_sell':int(data["token_sell"],16), 'token_buy':int(data["token_buy"],16),'nonce': int(data["nonce"]),'expiration_timestamp':int(data["expiration_timestamp"]),'fee_limit':int(data["fee_limit"]),'fee_token':int(data["fee_token"],16),'fee_vault_id':int(data["fee_vault_id"],10)}
    transfer_msg = get_limit_order_msg_with_fee(**data)
    r, s = sign(transfer_msg, int(private_key,16))
    return hex(r), hex(s)

def get_signature_local(data):
    private_key = data["sender_private_key"]
    data = {'amount': int(data["amount"]), 'nonce': int(data["nonce"]),
            'sender_vault_id': int(data["sender_vault_id"]),
            'token': int(data["asset_id"],16),
            'receiver_vault_id': int(data["receiver_vault_id"]),
            'receiver_public_key': int(data["receiver"],16),
            'expiration_timestamp': int(data["expiration_timestamp"])}
    transfer_msg = get_transfer_msg(**data)
    r, s = sign(transfer_msg, int(private_key,16))
    return hex(r), hex(s)

if __name__ == "__main__":
    stark_private_key = 'a7b68cf2ee72b2a0789914daa8ae928aec21b6b0bf020e394833f4c732d99d'
    stark_key = private_to_stark_key(int(stark_private_key,16))
    print(hex(stark_key))


from typing import Optional

from .signature import FIELD_PRIME, pedersen_hash


def get_msg(
        instruction_type: int, vault0: int, vault1: int, amount0: int, amount1: int, token0: int,
        token1_or_pub_key: int, nonce: int, expiration_timestamp: int,
        hash=pedersen_hash, condition: Optional[int] = None) -> int:
    """
    Creates a message to sign on.
    """
    packed_message = instruction_type
    packed_message = packed_message * 2**31 + vault0
    packed_message = packed_message * 2**31 + vault1
    packed_message = packed_message * 2**63 + amount0
    packed_message = packed_message * 2**63 + amount1
    packed_message = packed_message * 2**31 + nonce
    packed_message = packed_message * 2**22 + expiration_timestamp
    if condition is not None:
        # A message representing a conditional transfer. The condition is interpreted by the
        # application.
        return hash(hash(hash(token0, token1_or_pub_key), condition), packed_message)

    return hash(hash(token0, token1_or_pub_key), packed_message)


def get_limit_order_msg(
        vault_sell: int, vault_buy: int, amount_sell: int, amount_buy: int, token_sell: int,
        token_buy: int, nonce: int, expiration_timestamp: int,
        hash=pedersen_hash) -> int:
    """
    party_a sells amount_sell coins of token_sell from vault_sell.
    party_a buys amount_buy coins of token_buy into vault_buy.
    """
    assert 0 <= vault_sell < 2**31
    assert 0 <= vault_buy < 2**31
    assert 0 <= amount_sell < 2**63
    assert 0 <= amount_buy < 2**63
    assert 0 <= token_sell < FIELD_PRIME
    assert 0 <= token_buy < FIELD_PRIME
    assert 0 <= nonce < 2**31
    assert 0 <= expiration_timestamp < 2**22

    instruction_type = 0
    return get_msg(
        instruction_type, vault_sell, vault_buy, amount_sell, amount_buy, token_sell, token_buy,
        nonce, expiration_timestamp, hash=hash)


def get_transfer_msg(
        amount: int, nonce: int, sender_vault_id: int, token: int, receiver_vault_id: int,
        receiver_public_key: int, expiration_timestamp: int,
        hash=pedersen_hash, condition: Optional[int] = None) -> int:
    """
    Transfer `amount` of `token` from `sender_vault_id` to `receiver_vault_id`.
    The transfer is conditional only if `condition` is given.
    """
    assert 0 <= sender_vault_id < 2**31
    assert 0 <= receiver_vault_id < 2**31
    assert 0 <= amount < 2**63
    assert 0 <= token < FIELD_PRIME
    assert 0 <= receiver_public_key < FIELD_PRIME
    assert 0 <= nonce < 2**31
    assert 0 <= expiration_timestamp < 2**22

    TRANSFER = 1
    CONDITIONAL_TRANSFER = 2
    instruction_type = CONDITIONAL_TRANSFER if condition else TRANSFER
    assert condition is None or 0 <= condition < FIELD_PRIME

    return get_msg(
        instruction_type, sender_vault_id, receiver_vault_id, amount, 0, token, receiver_public_key,
        nonce, expiration_timestamp, hash=hash, condition=condition)


def get_limit_order_msg_with_fee(
        vault_sell: int, vault_buy: int, amount_sell: int, amount_buy: int, token_sell: int,
        token_buy: int, nonce: int, expiration_timestamp: int, fee_token:int, fee_limit: int,fee_vault_id: int,
        hash=pedersen_hash) -> int:
    """
    party_a sells amount_sell coins of token_sell from vault_sell.
    party_a buys amount_buy coins of token_buy into vault_buy.
    """
    assert 0 <= vault_sell < 2**31
    assert 0 <= vault_buy < 2**31
    assert 0 <= amount_sell < 2**63
    assert 0 <= amount_buy < 2**63
    assert 0 <= token_sell < FIELD_PRIME
    assert 0 <= token_buy < FIELD_PRIME
    assert 0 <= nonce < 2**31
    assert 0 <= expiration_timestamp < 2**22

    instruction_type = 3
    return get_msg_with_fee(
        instruction_type, vault_sell, vault_buy, amount_sell, amount_buy, token_sell, token_buy,
        nonce, expiration_timestamp,fee_limit,fee_token,fee_vault_id,  hash=hash)

def get_msg_with_fee(instruction_type: int, vault_sell: int, vault_buy: int, amount_sell: int, amount_buy: int, token_sell: int,
        token_buy: int, nonce: int, expiration_timestamp: int, fee_limit : int, fee_token : int, fee_vault_id : int,
        hash=pedersen_hash, condition: Optional[int] = None) -> int:
    packedMessage1 = amount_sell
    packedMessage1 = packedMessage1 * 2**64 + amount_buy
    packedMessage1 = packedMessage1 * 2**64 + fee_limit
    packedMessage1 = packedMessage1 * 2**32 + nonce

    packedMessage2 = instruction_type
    packedMessage2 = packedMessage2 * 2**64 + fee_vault_id
    packedMessage2 = packedMessage2 * 2**64 + vault_sell
    packedMessage2 = packedMessage2 * 2**64 + vault_buy
    packedMessage2 = packedMessage2 * 2**32 + expiration_timestamp
    packedMessage2 = packedMessage2 * 2**17

    tmpHash = hash(hash(token_sell, token_buy), fee_token)
    msgHash = hash(hash(tmpHash, packedMessage1), packedMessage2)
    return msgHash



import requests
import logging

from .settings import REDDIO_ENDPOINT_TESTNET, REDDIO_ENDPOINT_MAINNET

from .http_utils import request
from .starkex_utils import get_signature_local,get_asset_id

class Reddio(object):
    def __init__(self, env="testnet"):
        self.env = env
        if self.env == "testnet":
            self.endpoint = REDDIO_ENDPOINT_TESTNET
        elif self.env == "mainnet":
            self.endpoint = REDDIO_ENDPOINT_MAINNET

    def get_vault_id(self, stark_key, assetid):
        data = {"asset_id":assetid, "stark_keys": stark_key}
        headers = {'Content-Type': 'application/json'}
        r = requests.get(url = '%s/v1/vaults' % self.endpoint, params = data, headers = headers)
        vault_id = r.json()['data']['vault_ids'][0]
        return vault_id

    def get_order_info(self, stark_key, contract1Type, contract1Address, contract1TokenID, contract2Type, contract2Address, contract2TokenID):
        uri = '/v1/order/info' + '?stark_key=' + str(stark_key) + '&contract1=' + str(contract1Type) + ':' + str(contract1Address) + ':' + str(contract1TokenID) + '&contract2=' + str(contract2Type) + ':' + str(contract2Address) + ':' + str(contract2TokenID)
        url = self.endpoint + uri
        headers = {'Content-Type': 'application/json'}
        r = requests.get(url, headers = headers)
        return r.json()['data']


    def get_balances(self, stark_key, page=1, limit=10):
        uri = '/v1/balances' + '?stark_key=' + str(stark_key) + '&page=' + str(page) + '&limit=' + str(limit)
        url = self.endpoint + uri
        headers = {'Content-Type': 'application/json', "User-Agent":"ReddioFrame"}
        r = requests.get(url, headers = headers)
        return r.json()["data"]["list"]

    def get_orders(self, contract_address):
        uri = '/v1/orders' + '?contract_address=' + str(contract_address)
        url = self.endpoint + uri
        headers = {'Content-Type': 'application/json'}
        r = requests.get(url, headers = headers)
        order_list = []
        for order in r.json()["data"]["list"]:
            if order["stark_key"] != "":
                order_list.append(order)
        return order_list

    def list_nft_by_user(self, stark_key, contract_address):
        uri = '/v1/balances' + '?stark_key=' + str(stark_key) + '&contract_address=' + str(contract_address) + '&limit=1000'
        url = self.endpoint + uri
        headers = {'Content-Type': 'application/json'}
        r = requests.get(url, headers = headers)
        return r.json()["data"]["list"]

    def list_user_nfts(self, stark_key, contract_address):
        nfts = self.list_nft_by_user(stark_key, contract_address)
        nft_list = []
        for nft in nfts:
            if nft["balance_available"] > 0 and int(nft["token_id"]) > 668:
                nft_list.append(nft)
        return nft_list

    def get_nonce(self, stark_key):
        """
        Gets the nonce for a given stark key.
        """
        data = {"stark_key": stark_key}
        headers = {'Content-Type': 'application/json'}
        r = requests.get(url = '%s/v1/nonce' % self.endpoint , params = data, headers = headers)
        return r.json()['data']['nonce']
    
    def transferNFT(self, stark_private_key, sender_starkkey, receiver, token_type, contract, tokenID, expiration_timestamp=4194303):
        quantum = 1
        asset_id =  hex(get_asset_id(token_type, contract, quantum, int(tokenID)))
        vault_id = self.get_vault_id(sender_starkkey, asset_id)
        receiver_vault_id = self.get_vault_id(receiver, asset_id)
        nonce = self.get_nonce(sender_starkkey)
        data = {}
        data['asset_id'] = str(asset_id)
        data['receiver_vault_id'] = str(receiver_vault_id)
        data['sender_vault_id'] = str(vault_id)
        data['nonce'] = str(nonce)
        data['sender_private_key'] = str(stark_private_key)
        data['amount'] = "1"
        data['receiver'] = str(receiver)
        data['expiration_timestamp'] = str(expiration_timestamp)
        data['stark_key'] = str(sender_starkkey)

        transfer_data = get_transfer_data(data)

        url = self.endpoint + '/v1/transfers'
        headers = {'Content-Type': 'application/json'}
        x = request(url, transfer_data, headers)
        return x.json()

    def withdrawNFT(self, stark_private_key, sender_starkkey, receiver, token_type, contract, tokenID, expiration_timestamp=4194303):
        quantum = 1
        asset_id =  hex(get_asset_id(token_type, contract, quantum, int(tokenID)))
        vault_id = self.get_vault_id(sender_starkkey, asset_id)
        receiver_vault_id = self.get_vault_id(receiver, asset_id)
        nonce = self.get_nonce(sender_starkkey)
        data = {}
        data['asset_id'] = str(asset_id)
        data['receiver_vault_id'] = str(receiver_vault_id)
        data['sender_vault_id'] = str(vault_id)
        data['nonce'] = str(nonce)
        data['sender_private_key'] = str(stark_private_key)
        data['amount'] = "1"
        data['receiver'] = str(receiver)
        data['expiration_timestamp'] = str(expiration_timestamp)
        data['stark_key'] = str(sender_starkkey)

        transfer_data = get_transfer_data(data)

        url = self.endpoint + '/v1/withdrawalto'
        headers = {'Content-Type': 'application/json'}
        x = request(url, transfer_data, headers)
        return x.json()

def get_transfer_data(data):
    r, s = get_signature_local(data)
    original_transfer_data = {}
    original_transfer_data['amount'] = data['amount']
    original_transfer_data['nonce'] = int(data['nonce'])
    original_transfer_data['vault_id'] = data['sender_vault_id']
    original_transfer_data['receiver_vault_id'] = data['receiver_vault_id']
    original_transfer_data['stark_key'] = data["stark_key"]
    original_transfer_data['receiver'] = data["receiver"]
    original_transfer_data['signature'] = {"r": r, "s": s}
    original_transfer_data['asset_id'] = data['asset_id']
    original_transfer_data['expiration_timestamp'] = int(data['expiration_timestamp'])
    return original_transfer_data
if __name__ == "__main__":
    obj = Reddio("testnet")
    # print(obj.get_balances("0x6ecaebbe5b9486472d964217e5470380782823bb0d865240ba916d01636310a"))

    # print(obj.transferNFT('private_key', '0x6ecaebbe5b9486472d964217e5470380782823bb0d865240ba916d01636310a', '0x1ada455b26b246260b7fd876429289639d7a0ce5fe295ff2355bd4f4da55e2', 'ERC721', '0x941661Bd1134DC7cc3D107BF006B8631F6E65Ad5', '618'))
    # print(obj.withdrawNFT('private_key', '0x6ecaebbe5b9486472d964217e5470380782823bb0d865240ba916d01636310a', '0xffc882996cFAB2C8B9983394E09bb025a98e52bc', 'ERC721', '0x941661Bd1134DC7cc3D107BF006B8631F6E65Ad5', '663'))
    # print(obj.get_balances("0x1ada455b26b246260b7fd876429289639d7a0ce5fe295ff2355bd4f4da55e2"))